/**
 * @fileoverview typings for "eslint/config" module
 * @author Nicholas C. Zakas
 */

import { defineConfig, globalIgnores } from "@eslint/config-helpers";

export { defineConfig, globalIgnores };
